# boggle_board.py
"""
Library to build a representation of the Boggle board and provide the solver
with a simple map to navigate between neighboring tiles on the board.
"""
from math import floor, sqrt
from re import sub


class BoggleBoard():
    """
    Represents a view of the letter tiles for the puzzle to be solved.
    """

    def __init__(self, tiles=None):
        """
        Initialise the empty board and fill it with letter tiles if provided.
        """
        self._columns = None
        self._neighbors = None
        self._rangeLimit = None
        self._rows = None
        self._tiles = None
        if tiles:
            self.createBoard(tiles)

    @property
    def columns(self):
        """
        The number of columns in the grid that makes up the board.

        The property is populated when the `createBoard` method is called.

        >>> ex = BoggleBoard('ABCD EFGH IJKL MNOP')
        >>> ex.columns
        4
        """
        return self._columns

    @property
    def neighbors(self):
        """
        A dict structure in which the keys are the enumerated positions on the
        board and the values are the list of neighbors for that position.

        The property is populated when the `createBoard` method is called.

        >>> ex = BoggleBoard('A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P')
        >>> isinstance(ex.tiles, dict)
        True
        >>> len(ex.neighbors)
        16
        >>> ex.neighbors.keys()
        dict_keys([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16])
        >>> isinstance(ex.neighbors[1], list)
        True
        >>> ex.neighbors[1]
        [6, 2, 5]
        >>> ex.neighbors[2]
        [1, 3, 5, 6, 7]
        >>> ex.neighbors[3]
        [2, 4, 6, 7, 8]
        >>> ex.neighbors[4]
        [3, 7, 8]
        >>> ex.neighbors[5]
        [1, 2, 6, 9, 10]
        >>> ex.neighbors[6]
        [1, 2, 3, 5, 7, 9, 10, 11]
        >>> ex.neighbors[7]
        [2, 3, 4, 6, 8, 10, 11, 12]
        >>> ex.neighbors[8]
        [3, 4, 7, 11, 12]
        >>> ex.neighbors[9]
        [5, 6, 10, 13, 14]
        >>> ex.neighbors[10]
        [5, 6, 7, 9, 11, 13, 14, 15]
        >>> ex.neighbors[11]
        [6, 7, 8, 10, 12, 14, 15, 16]
        >>> ex.neighbors[12]
        [7, 8, 11, 15, 16]
        >>> ex.neighbors[13]
        [9, 10, 14]
        >>> ex.neighbors[14]
        [9, 10, 11, 13, 15]
        >>> ex.neighbors[15]
        [10, 11, 12, 14, 16]
        >>> ex.neighbors[16]
        [11, 12, 15]
        """
        return self._neighbors

    @property
    def rangeLimit(self):
        """
        The upper limit for the range of the tiles on the board.

        The property is populated when the `createBoard` method is called.

        >>> ex = BoggleBoard('A B C D E F G H I J K L M N O P')
        >>> ex.rangeLimit
        17
        """
        return self._rangeLimit

    @property
    def rows(self):
        """
        The number of rows in the grid that makes up the board.

        The property is populated when the `createBoard` method is called.

        >>> ex = BoggleBoard('ABCDEFGHIJKLMNOP')
        >>> ex.rows
        4
        """
        return self._rows

    @property
    def tiles(self):
        """
        A dict structure in which the keys are the enumerated positions on the
        board and the values content of the letter tile.

        The property is populated when the `createBoard` method is called.

        >>> ex = BoggleBoard('ABCDEFGHIJKLMNOP')
        >>> isinstance(ex.tiles, dict)
        True
        >>> len(ex.tiles)
        16
        >>> ex.tiles.keys()
        dict_keys([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16])
        >>> ex.tiles[1]
        'A'
        >>> ex.tiles[8]
        'H'
        >>> ex.tiles[16]
        'P'
        """
        return self._tiles

    @tiles.setter
    def tiles(self, new_tiles: str):
        """
        Repopulates the board with new tiles.
        """
        self.createBoard(new_tiles)

    def createBoard(self, tiles):
        """
        Creates the board and determines its properties by parsing the string
        of tile values.

        Args:
            tiles: String of characters to be placed on the board.

        Any fragment will be split into individual characters if it is more
        than two characters in length.

        The string can be in a number of formats and mixed formats are handled:
        >>> ex_1 = BoggleBoard()
        >>> ex_1.createBoard('ABCDEFGH')
        >>> len(ex_1.tiles)
        8
        >>> ex_1.tiles
        {1: 'A', 2: 'B', 3: 'C', 4: 'D', 5: 'E', 6: 'F', 7: 'G', 8: 'H'}
        >>> ex_2 = BoggleBoard()
        >>> ex_2.createBoard('A B C D E F')
        >>> len(ex_2.tiles)
        6
        >>> ex_2.tiles
        {1: 'A', 2: 'B', 3: 'C', 4: 'D', 5: 'E', 6: 'F'}
        >>> ex_3 = BoggleBoard()
        >>> ex_3.createBoard('Z, X, Y, W, V, U')
        >>> len(ex_3.tiles)
        6
        >>> ex_3.tiles
        {1: 'Z', 2: 'X', 3: 'Y', 4: 'W', 5: 'V', 6: 'U'}
        >>> ex_4 = BoggleBoard()
        >>> ex_4.createBoard('A, B CH TH E,F')
        >>> len(ex_4.tiles)
        6
        >>> ex_4.tiles
        {1: 'A', 2: 'B', 3: 'CH', 4: 'TH', 5: 'E', 6: 'F'}
        """
        if tiles.isalpha():
            _list = [ch for ch in tiles]
        else:
            _list = []
            for el in (sub(r'\W+', ' ', tiles)).split():
                _list.extend([ch for ch in el] if (len(el) > 2) else [el])
        self._tiles = {pos: tile.upper()
                       for pos, tile in enumerate(_list, 1)}
        num_tiles = len(self._tiles)
        cols = floor(sqrt(num_tiles))
        while num_tiles % cols is not 0:
            cols -= 1
        # TODO: What to do if we are left with a single column?
        # i.e. num_tiles is a prime number
        self._rows = num_tiles // cols
        self._columns = cols
        self._rangeLimit = num_tiles + 1
        self._neighbors = {p: [n for n in self.positionNeighbors(p)]
                           for p in range(1, self._rangeLimit)}

    def positionNeighbors(self, position: int):
        """
        Returns a list of valid neighbors for the provided board position.
        Raises an IndexError if the position is not in the enumerated board.

        Args:
            position: int in the range of 1 to the total number of tiles.

        This method is called within the `createBoard` method to populate the
        neighbors property.

        Knowing the number of columns in a grid and the total number of
        positions it contains allows the neighbors for a given position to be
        determined easily.

        Using the following grid as an example:
        +---+---+---+---+
        |1  |2  |3  |4  |
        | A | B | C | D |
        +---+---+---+---+
        |5  |6  |7  |8  |
        | E | F | G | H |
        +---+---+---+---+
        |9  |10 |11 |12 |
        | I | J | K | L |
        +---+---+---+---+
        |13 |14 |15 |16 |
        | M | N | O | P |
        +---+---+---+---+

        Number of columns = 4
        Number of positions = 16

        Each position in the grid potentially has 8 neighbors as illustrated
        in the grid below.

        +---+---+---+
        |0  |1  |2  |
        |   |   |   |
        +---+---+---+
        |3  |4  |5  |
        |   | X |   |
        +---+---+---+
        |6  |7  |8  |
        |   |   |   |
        +---+---+---+

        The list of possible neighbors for 'X' is [0, 1, 2, 3, 5, 6, 7, 8].

        The position value for each possible neighbor in the list can be
        determined by applying the appropriate calculation from the following:

        0 : (current position) - (number of columns + 1)
        1 : (current position) - (number of columns)
        2 : (current position) - (number of columns - 1)
        3 : (current position) - 1
        5 : (current position) + 1
        6 : (current position) + (number of columns - 1)
        7 : (current position) + (number of columns)
        8 : (current position) + (number of columns + 1)

        The positions on the edge of the board will have fewer valid neighbors.

        The enumerated board positions simplifies the checks that need to be
        performed to find a position's neighbors.

        Board properties for eliminating neighbors for each edge case:

        Top    : (current position) - (number of columns) < 1
        Right  : (current position) % (number of columns) = 0
        Left   : (current position) % (number of columns) = 1
        Bottom : (current position) + (number of columns) > (positions count)

        Applying these rules will determine a list of neighbors that are not
        valid for the given position. The list may contain duplicate items,
        these are removed by creating a set from the list.

        The set of valid neighbors is the set of all neighbors less the set
        of invalid neighbors.

        >>> ex = BoggleBoard('A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P')
        >>> ex.positionNeighbors(2)
        [1, 3, 5, 6, 7]
        >>> ex.positionNeighbors(4)
        [3, 7, 8]
        >>> ex.positionNeighbors(5)
        [1, 2, 6, 9, 10]
        >>> ex.positionNeighbors(7)
        [2, 3, 4, 6, 8, 10, 11, 12]
        >>> ex.positionNeighbors(12)
        [7, 8, 11, 15, 16]
        >>> ex.positionNeighbors(14)
        [9, 10, 11, 13, 15]
        """
        if 1 <= position < self._rangeLimit:
            cols = self._columns
            possible_neighbors = [-(cols + 1), -(cols), -(cols - 1),
                                  -1, 0, +1,
                                  (cols - 1), cols, (cols + 1)]
            invalid_neighbors = [4]
            if position - cols < 1:
                invalid_neighbors.extend([0, 1, 2])
            if (position % cols) == 0:
                invalid_neighbors.extend([2, 5, 8])
            if (position % cols) == 1:
                invalid_neighbors.extend([0, 3, 6])
            if position + cols >= self._rangeLimit:
                invalid_neighbors.extend([6, 7, 8])
            valid_neighbors = set(range(9)).difference(set(invalid_neighbors))
            return [(position + possible_neighbors[n])
                    for n in valid_neighbors]
        raise IndexError



# boggle_dice.py
"""
Library of classes for the magic Boggle dice and their container.
"""
from collections import Counter
import os
from random import choice, sample, seed
from re import sub


class Dice():
    """
    Represents the N sided dice.
    """

    def __init__(self, tokens=None):
        """
        Initialise the dice with tokens if they are provided.

        >>> ex = Dice()
        >>> ex.tiles is None
        True
        >>> ex.lastRoll is None
        True
        """
        self._tiles = None
        self._lastRoll = None
        if tokens is not None:
            self.makeDice(tokens)

    @property
    def tiles(self):
        """
        A tuple representing the tiles which make up the sides of the dice.

        >>> ex = Dice('ABC')
        >>> isinstance(ex.tiles, tuple)
        True
        >>> len(ex.tiles)
        3
        >>> ex.tiles
        ('A', 'B', 'C')
        """
        return self._tiles

    @property
    def lastRoll(self):
        """
        The result of the last roll of the dice.

        >>> ex = Dice('AAA')
        >>> isinstance(ex.lastRoll, str)
        True
        >>> ex.lastRoll
        'A'
        """
        return self._lastRoll

    def makeDice(self, tokens: str):
        """
        Make a dice from a tokens string and return the first roll result.

        Args:
            tokens: String of letters to appear on the dice.

        The dice is created by splitting the alphabetical elements of the
        token string into a tuple. If the token string is contiguous text
        it will be split into individual characters.

        >>> ex_1 = Dice()
        >>> ex_1.makeDice('AAA')
        'A'
        >>> ex_1.tiles
        ('A', 'A', 'A')
        >>> ex_2 = Dice()
        >>> ex_2.makeDice('B B B B B BB')
        'B'
        >>> ex_2.tiles
        ('B', 'B', 'B', 'B', 'B', 'BB')
        >>> ex_3 = Dice()
        >>> ex_3.makeDice('C, CC, C, C, C, C')
        'C'
        >>> ex_3.tiles
        ('C', 'CC', 'C', 'C', 'C', 'C')
        """
        seed()
        if tokens.isalpha():
            _tiles = tuple([t for t in tokens])
        else:
            _tiles = tuple((sub(r'\W+', ' ', tokens)).split())
        self._tiles = _tiles
        return self.roll()

    def roll(self):
        """
        Rolls the dice by randomly choosing an element and returns the result.

        >>> ex = Dice('B B B B B B')
        >>> ex.lastRoll
        'B'
        """
        if self._tiles is not None:
            self._lastRoll = choice(self._tiles)
        return self._lastRoll

    def __iter__(self):
        return self

    def __next__(self):
        """
        Returns next dice roll result.

        >>> ex = Dice('A A A A A A')
        >>> next(ex)
        'A'
        """
        if self._tiles:
            return [self.roll()].pop(0)
        raise StopIteration


class BoggleDice():
    """
    Represents the container which holds the game dice.
    """

    def __init__(self):
        """
        Initialise an empty dice container.

        >>> bd = BoggleDice()
        >>> isinstance(bd.dice, list)
        True
        >>> len(bd.dice)
        0
        >>> bd.letters
        Counter()
        >>> bd.lastRoll is None
        True
        """
        self._container = []
        self._letters = Counter()
        self._rolled = None

    def addDiceFromTokensString(self, tokens: str):
        """
        Adds a dice to the container from a string of tokens.

        Args:
            tokens: String of characters to be be converted into dice tiles.

        The string which is converted to uppercase before the dice is created.

        >>> bd = BoggleDice()
        >>> bd.addDiceFromTokensString('AAA')
        >>> len(bd.dice)
        1
        >>> isinstance(bd.dice[0], Dice)
        True
        """
        dice = Dice(tokens.upper())
        self._container.extend([dice])
        self._letters.update(set(dice.tiles))

    @property
    def dice(self):
        """
        The list of Dice objects which make up the container.
        """
        return self._container

    @property
    def letters(self):
        """
        A Counter object which is records the set of possible letters from
        each dice added to the container.

        This can be useful data for filtering a word dictionary for use
        with Boggle.
        """
        return self._letters

    @property
    def lastRoll(self):
        """
        A list containing the result of the container being shaken.
        """
        return self._rolled

    def loadFromFile(self, file_path: str):
        """
        Populates a dice container from the contents of a text file.

        Args:
            file_path: String containing the path to the text file

        Each line of the file is treated as a string of tokens for creating
        a new dice. The new dice is added to the container.
        """
        with open(file_path, 'r') as _file:
            for tokens in _file.readlines():
                self.addDiceFromTokensString(tokens.strip())

    def shake(self):
        """
        Shakes the container to randomize the dice inside.

        >>> bd = BoggleDice()
        >>> bd.addDiceFromTokensString('CCC')
        >>> bd.addDiceFromTokensString('C C C')
        >>> bd.addDiceFromTokensString('C, C, C')
        >>> bd.shake()
        ['C', 'C', 'C']
        """
        coll = self._container
        self._rolled = sample([next(d) for d in coll], len(coll))
        return self._rolled



# boggle_solver.py
"""
Library for solving a provided Boggle puzzle using a word dictionary.
"""
from boggled.boggle_board import BoggleBoard
from boggled.boggle_dice import BoggleDice
from boggled.boggle_words import BoggleWords, TrieNode


class BoggleSolver():
    """
    Represents the all singing and dancing Boggle puzzle solver.
    """

    def __init__(self, board, dictionary):
        """
        Initialise the solver with the provided board and dictionary.
        """
        self._board = board
        self._dictionary = dictionary.words
        self._minWordLength = dictionary.minLength
        self._prefixes = dictionary.prefixes
        self._found = dict()

    @property
    def board(self):
        """
        The board created from the provided letters.
        """
        return self._board

    @property
    def boardNeighbors(self):
        """
        Retrieve the neighbors grid from the BoggleBoard instance.
        """
        return self._board.neighbors

    @property
    def boardRangeLimit(self):
        """
        The upper range limit of the BoggleBoard instance.
        """
        return self._board.rangeLimit

    @property
    def boardTiles(self):
        """
        The letter tiles on the board.
        """
        return self._board.tiles

    @property
    def dictionary(self):
        """
        The Trie data structure used to find words on the board.
        """
        return self._dictionary

    @property
    def minWordLength(self):
        """
        The minimum length word to find when solving the board.
        """
        return self._minWordLength

    @property
    def prefixes(self):
        """
        The word prefixes found in the dictionary.
        """
        return self._prefixes

    @property
    def found(self):
        """
        A dict of tuples containing each word found as the key and a list of
        paths on the board where the word was found. There may be duplicate
        paths in the list as there can be mor than one way to make up the same
        word.

        For example in the following board fragment, the word 'BOOK' is found
        by following two different paths, 2->1->5->6 and 2->5->1->6.

        +---+---+---+---+
        |1  |2  |3  |4  |
        | O | B | Z | Q |
        +---+---+---+---+
        |5  |6  |7  |8  |
        | O | K | N | H |
        +---+---+---+---+

        In this case the list will contain two path tuples for the same word:

        {'BOOK': [(2, 1, 5, 6), (2, 5, 1, 6)]

        Each tuple in the list will always be unique.
        """
        return self._found

    @property
    def foundWords(self):
        """
        A sorted list of the words found on the board.
        """
        return sorted([word for word in self._found.keys()])

    def _addFoundWord(self, details: tuple):
        """
        Adds the details of the word found to the collection.

        Args:
            details: Tuple - the word and the path followed to create it.
        """
        word, path = details
        if self._found.get(word, None) is None:
            self._found[word] = [path]
        else:
            self._found[word].extend([path])

    def _getNode(self, prefix: str, start=None):
        """
        A helper method to retrieve the node from the dictionary's trie
        structure which corresponds to the prefix string value.
        """
        return self._dictionary.node(prefix, start)

    def _hasValidPrefix(self, prefix: str, path: list):
        """
        Indicates if the prefix string and the path used to construct it are
        valid.

        Args:
            prefix: String to match in the Trie structure.
            path: List of positions used to compose the string.

        Returns:
            Boolean indicating if the prefix meets the requirements.

        A prefix is valid if it contains the minimum word length number of
        letters and can be found in the prefixes dictionary.
        """
        return (len(path) <= self.minWordLength
                and prefix in self.prefixes)

    def _findPrefixPaths(self, prefix: str, path: list):
        """
        Yields each prefix found in the path as a tuple with three elements.

        Args:
            prefix: String to match in the Trie structure.
            path: List of neighbors used to compose the string.

        Returns:
            tuple(prefix, path, node)
            prefix: String containing the prefix.
            path: List of neighbors used to compose the string.
            node: TrieNode that corresponds to the last letter in the string.
        """
        minLength = self.minWordLength
        next_neighbors = self._findNeighborsToVisit(path)
        tiles = self.boardTiles
        if self._hasValidPrefix(prefix, path):
            yield (prefix, path, self.prefixes[prefix[:minLength]])
        for neighbor in next_neighbors:
            if len(path) < minLength:
                next_prefix = ''.join([prefix, tiles[neighbor]])
                next_path = path + [neighbor]
                for result in self._findPrefixPaths(next_prefix, next_path):
                    yield result

    def _findNeighborsToVisit(self, path):
        """
        Returns the list of neighbors not present in the path.

        Args:
            path: List of positions already visited.

        Returns:
            List of neighbor positions not visited.
        """
        position_neighbors = self.boardNeighbors[path[-1]]
        return [position for position in position_neighbors if position not in path]

    def _findWords(self, prefix: str, path: list, node: TrieNode=None):
        """
        A recursive method which finds all words in the dictionary beginning
        with the prefix string and extending from the path on the board.

        Args:
            prefix: String to match in the Trie structure.
            path: List of neighbors used to compose the string.
            node: TrieNode that corresponds to the last letter in the string.

        Algorithm:
        If the prefix string corresponds with a word in the dictionary:
            Add the tuple (prefix, path) to the list for words already found.
        For each neighboring position not yet visited:
            Make a new prefix by appending the position's value to the prefix.
            Make a new path by appending the position to the path.
            Call the method with the updated prefix and path values.
        """
        if node is None:
            node = self._getNode(prefix)
        if node is not None:
            if node.isWord:
                self._addFoundWord((prefix, tuple(path)))
            next_neighbors = self._findNeighborsToVisit(path)
            tiles = self.boardTiles
            for neighbor in next_neighbors:
                next_prefix = ''.join([prefix, tiles[neighbor]])
                next_path = path + [neighbor]
                next_node = self._getNode(tiles[neighbor], node)
                if next_node is not None:
                    self._findWords(next_prefix, next_path, next_node)

    def solve(self):
        """
        Solves the board in a two step process.

        First the list of word prefixes along with their neighbor path on the
        board is identified.

        Each prefix found is then used as a starting point for finding a
        dictionary word.
        """
        tiles = self.boardTiles
        paths = [path for position in range(1, self.boardRangeLimit)
                 for path in self._findPrefixPaths(tiles[position], [position])]
        for prefix, path, node in paths:
            self._findWords(prefix, path, node)



# boggle_words.py
"""
Library for building a word dictionary including prefixes for Boggle.
"""
from collections import UserDict
from re import sub


class NoneFieldDict(UserDict):
    """
    This dict subclass returns a None type instead of raising a KeyError.
    """

    def __missing__(self, key):
        "Returns the null object None."
        return None


class TrieNode:
    """
    Represents a Trie data structure used to build an efficient word search.

    The children container is implemented using the `NoneFieldDict`. It's use
    of the __missing__ method reduces the memory footprint of the Trie.

    Public attributes:
    - parent   : TrieNode - Link to parent node in the Trie.
    - children : NoneFieldDict of TrieNode - The key is the next letter.
    - isWord   : Boolean - Indicates this node is the last letter in a word.
    """

    def __init__(self, parent, value):
        """
        Initialise the TrieNode and link it with the parent node.
        """
        self._parent = parent
        self._children = NoneFieldDict()
        self._isWord = False
        if parent is not None:
            parent._children[value[0]] = self

    @property
    def parent(self):
        """
        Parent node in the Trie.
        TrieNode / None
        """
        return self._parent

    @property
    def children(self):
        """
        Children of the node.
        NoneFieldDict: key: character string, value: TrieNode
        """
        return self._children

    @property
    def isWord(self):
        """
        Indicates whether or not this node is the last letter in a word.
        """
        return self._isWord


class Trie:
    """
    Represents the root node of the Trie structure.

    Public attributes:
    root : TrieNode - The root element of the Trie.
    """

    def __init__(self):
        """
        Initialise the Trie with an empty node.
        """
        self._root = TrieNode(None, '')

    @property
    def root(self):
        """
        The root element of the Trie.
        """
        return self._root

    def node(self, prefix, start=None):
        """
        Returns the node which corresponds to the final character of the
        prefix.

        Returns None if the prefix is not present in the Trie.
        """
        node = start if start else self._root
        for ch in prefix:
            if node.children[ch] is not None:
                node = node.children[ch]
            else:
                node = None
                break
        return node


class BoggleWords():
    """
    Represents the words used in the game dictionary.

    Public attributes:
    minLength : int - The prefix string length.
    prefixes : dict - key: prefix string, value: TrieNode.
    words: Trie - The dictionary data structure.
    wordsRoot: TrieNode - The root node of the dictionary.
    """

    def __init__(self, minLength=3):
        """
        Initialise with a minimum word length of 3 if not specified.
        """
        self._minLength = minLength
        self._prefixes = dict()
        self._words = Trie()

    @property
    def minLength(self):
        """
        Word prefix length and minimum length word for the game.
        """
        return self._minLength

    @property
    def prefixes(self):
        """
        dict for dictionary word prefixes - key: prefix string, value: TrieNode.
        """
        return self._prefixes

    @property
    def words(self):
        """
        The dictionary Trie.
        """
        return self._words

    @property
    def wordsRoot(self):
        """
        TrieNode at the root of the dictionary.
        """
        return self._words.root

    def _cleanWord(self, word):
        """
        Returns word in uppercase with all non-alphabethic characters removed.

        Args:
            word: String of the word being added to the Trie.

        Returns:
            Uppercase string or False
        """
        cleaned = sub(r'[\W]', '', word)
        is_valid = cleaned.isalpha() and (len(cleaned) >= self.minLength)
        return cleaned.upper() if is_valid else False

    def _getWordStartNode(self, word):
        """
        Returns the node where the word insertion should begin.

        Looks ahead to see if word prefix is already present, this will allow
        the initial levels of the Trie to be skipped if the prefix is there.

        Args:
            word: String of the word being added to the Trie.

        Returns:
            tuple(start_node, word, prefix)
            start_node: TrieNode where to begin process.
            word: String to be added.
            prefix: String prefix if it's a new prefix, otherwise None.
        """
        prefix = word[:self._minLength]
        if (prefix not in self._prefixes):
            start_node = self.wordsRoot
        else:
            start_node = self.prefixes[prefix]
            prefix = None
            word = word[self._minLength:]
        return (start_node, word, prefix)

    def iteratorPopulateTrie(self, word_iter):
        """
        Populates the Trie structure from an iterator of words.
        """
        for word in word_iter:
            cleaned_word = self._cleanWord(word)
            if isinstance(cleaned_word, str):
                node, word_part, prefix = self._getWordStartNode(cleaned_word)
                for index, letter in enumerate(word_part, 1):
                    next_node = node.children[letter]
                    if next_node is None:
                        next_node = TrieNode(node, letter)
                    node = next_node
                    if prefix is not None and index == len(prefix):
                        self.prefixes[prefix] = node
                node._isWord = True

    def loadFromFile(self, file_path):
        """
        Reads the list of words from a file and passes the list to the
        iteratorPopulateTrie() method.

        Args:
            file_path: Path to the text file.
        """
        with open(file_path, 'r') as words_file:
            self.iteratorPopulateTrie(words_file.readlines())



from setuptools import setup, find_packages

long_description = """
A suite of classes that model the elements of the Hasbro game. Boards of 
different sizes can be randomly generated from a collection of  dice.
Given a list of words, a puzzle can be solved by finding all of the words on
the Boggle board.

BoggleDice - The game dice - you can roll them.
BoggleBoard - The board, place the dice on the board and shake it!
BoggleWords - A dictionary for finding words.
BoggleSolver - Finds words from the dictionary that are on the board.

The official game letters and a dictionary file are NOT included.
"""

setup(
    name="boggled",
    description="Simple Boggle game solver and generator.",
    version="1.0.0",
    author="Dan Nagle",
    author_email="d.a.nagle@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/danag/boggled",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
